#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>
#include <cctype>
#include <cmath>
#include "tnt.h"
#include "jama_qr.h"

using namespace std;
using namespace TNT;
using namespace JAMA;


// function read_data
bool read_data(char* filename, int& no_entries, vector<string>& buffer)
{
    ifstream input(filename);
    if (!input) {
        cerr << "Couldn't open file " << filename << "!" << endl;
        return false;
    }
    
    string temp;
    while (getline(input, temp) ) {
        if (temp.substr(0, 2) == "/*" || temp.substr(0, 2) == "//" )
            continue;
        if (isdigit(temp[0]) ) continue;
        no_entries++;
        buffer.push_back(temp);
    }
    
    return true;
}


// function parse_data
bool parse_data(int no_entries, vector<string>& buffer, Array2D<double>& A, Array1D<double>& b)
{
    istringstream ins;
    
    string temp;
    double x, y, z, qx, qy, qz, qw, quality;
    int valid;
    
    for (int i = 0; i < no_entries; i++) {
        ins.str(buffer[i]);
        
        ins >> temp >> x >> y >> z >> qx >> qy >> qz >> qw >> quality >> valid;
        
        b[i*3 + 0] = -x;
        b[i*3 + 1] = -y;
        b[i*3 + 2] = -z;
        
        A[i*3 + 0][0] = 1.0 - 2.0 * (qy*qy + qz*qz);
        A[i*3 + 1][1] = 1.0 - 2.0 * (qx*qx + qz*qz);
        A[i*3 + 2][2] = 1.0 - 2.0 * (qx*qx + qy*qy);
        A[i*3 + 1][0] = 2.0 * (qx*qy + qz*qw);
        A[i*3 + 2][0] = 2.0 * (qx*qz - qy*qw);
        A[i*3 + 0][1] = 2.0 * (qx*qy - qz*qw);
        A[i*3 + 2][1] = 2.0 * (qy*qz + qx*qw);
        A[i*3 + 0][2] = 2.0 * (qx*qz + qy*qw);
        A[i*3 + 1][2] = 2.0 * (qy*qz - qx*qw);
        
        A[i*3 + 0][3] = -1.0; A[i*3 + 0][4] =  0.0; A[i*3 + 0][5] =  0.0;
        A[i*3 + 1][3] =  0.0; A[i*3 + 1][4] = -1.0; A[i*3 + 1][5] =  0.0;
        A[i*3 + 2][3] =  0.0; A[i*3 + 2][4] =  0.0; A[i*3 + 2][5] = -1.0;
    }
    
    return true;
}


double calc_solution(const int no_entries, const Array2D<double>& A, const Array1D<double>& b,
                   Array1D<double>& object_coord, Array1D<double>& world_coord)
{
    Array1D<double> x(6, 1);
    QR<double> decomp_A(A);
    x = decomp_A.solve(b);
    
    object_coord[0] = x[0];
    object_coord[1] = x[1];
    object_coord[2] = x[2];
    
    world_coord[0]  = x[3];
    world_coord[1]  = x[4];
    world_coord[2]  = x[5];
    
    Array1D<double> c(no_entries * 3);
    for (int i = 0; i < no_entries; i++) {
        c[i*3 + 0] = (A[i*3+0][0] * x[0] + A[i*3+0][1] * x[1] + A[i*3+0][2] * x[2]
                      + A[i*3+0][3] * x[3] + A[i*3+0][4] * x[4] + A[i*3+0][5] * x[5]) - b[i*3+0];
    }
    
    double residual = 0.0;
    for (int i = 0; i < no_entries; i++) 
        residual += sqrt( c[i*3+0]*c[i*3+0] + c[i*3+1]*c[i*3+1] + c[i*3+2]*c[i*3+2]);
    residual = residual / no_entries;
    
    return residual;
}


// function main
int main(int argc, char* argv[])
{
    if (argc != 2) {
        cerr << "Usage: " << argv[0] << " <file>" << endl;
        return 1;
    }
    
    int no_entries = 0;
    vector<string> buffer;
    
    bool success = read_data(argv[1], no_entries, buffer);
    
    if (!success)
        return 1;
        
    cout << "Read " << no_entries << " entries from " << argv[1] << "." << endl;
    
    Array2D<double> A(no_entries * 3, 6);
    Array1D<double> b(no_entries * 3);
    
    parse_data(no_entries, buffer, A, b);
    
    cout << "Parsed " << no_entries << " lines." << endl;
    
    Array1D<double> object_coord(3, 1);
    Array1D<double> world_coord(3, 1);
    
    double residual = calc_solution(no_entries, A, b, object_coord, world_coord);
    
    cout << "Result: Object coordinates = (" << object_coord[0] << ", " << object_coord[1]
         << ", " << object_coord[2] << ")" << endl;
    cout << "        World coordinates  = (" << world_coord[0] << ", " << world_coord[1]
         << ", " << world_coord[2] << ")" << endl;
    cout << "        Error = " << residual << endl << endl;
    
    return 0;
}


// end of calib.cpp
